/**
 * Copyright (C) 2009-2015 Tatjana (Tanja) Kostic
 * <p>
 * This file belongs to jCleanCim, a tool supporting tasks of UML model managers for IEC TC57 CIM
 * and 61850 models.
 * <p>
 * This program is free software: you can redistribute it and/or modify it under the terms of the
 * GNU Lesser General Public License version 3.
 * <p>
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details. You should have received a copy of the GNU Lesser
 * General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

package org.tanjakostic.jcleancim.docgen.collector.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Set;

import org.tanjakostic.jcleancim.docgen.collector.DocIDHelper;
import org.tanjakostic.jcleancim.docgen.collector.DocgenConfig;
import org.tanjakostic.jcleancim.docgen.collector.RawData;
import org.tanjakostic.jcleancim.docgen.collector.TableSpec;
import org.tanjakostic.jcleancim.docgen.collector.WAX;
import org.tanjakostic.jcleancim.docgen.collector.impl.ag.AttributeGroup;
import org.tanjakostic.jcleancim.model.PresenceCondition;
import org.tanjakostic.jcleancim.model.UML;
import org.tanjakostic.jcleancim.model.UmlAttribute;
import org.tanjakostic.jcleancim.model.UmlClass;
import org.tanjakostic.jcleancim.model.UmlModel;
import org.tanjakostic.jcleancim.model.UmlObject;
import org.tanjakostic.jcleancim.util.Util;

/**
 * Holds the utility method to filter groups which may stay with no members after applying
 * configuration filters (e.g., a class actually inherits from another class, but another class is
 * informative, and we don't want informative stuff printed), and some common formatting.
 *
 * @author tatjana.kostic@ieee.org
 * @version $Id: Attributes61850Doc.java 2728 2015-05-14 20:00:03Z CHTAKOS $
 */
public abstract class Attributes61850Doc extends AbstractPropertiesDoc {

	/** String to print in 61850-7-3 if data attribute (of DA or CDC) is primitive. */

	/** String to print in 61860-7-4 when data object is transient. */
	protected static final String TRANSIENT = "T";

	public static final String INTRO_FMT = " shows all attributes of %s.";
	public static final String CAPTION_FMT = "Attributes of %s";

	/**
	 * Constructor.
	 */
	protected Attributes61850Doc(DocgenConfig docgenCfg, UmlObject object, String what,
			String headingText, String introText, String captionText, TableSpec colSpec,
			String tableName) {
		super(docgenCfg, object, what, null, null, true, headingText, introText, captionText,
				colSpec, tableName);
	}

	/**
	 * Returns non-empty groups, retained from <code>groups</code> after applying the filters set in
	 * configuration.
	 */
	protected final Collection<AttributeGroup> filterGroups(Collection<AttributeGroup> groups) {
		List<AttributeGroup> result = new ArrayList<AttributeGroup>();
		for (AttributeGroup group : groups) {
			List<UmlAttribute> filteredNatives = new ArrayList<UmlAttribute>();
			for (UmlAttribute a : group.getNativeAttributes()) {
				if (!toSkip(a)) {
					filteredNatives.add(a);
				}
			}
			List<UmlAttribute> filteredInheriteds = new ArrayList<UmlAttribute>();
			for (UmlAttribute a : group.getInheritedAttributes()) {
				if (!toSkipInherited(a)) {
					filteredInheriteds.add(a);
				}
			}
			if (!filteredNatives.isEmpty() || !filteredInheriteds.isEmpty()) {
				result.add(new AttributeGroup(group.getAgSpec(), filteredNatives,
						filteredInheriteds));
			}
		}
		return result;
	}

	/**
	 * Fills <code>outRawData</code> with value for
	 * {@value org.tanjakostic.jcleancim.docgen.collector.WAX#A_deducedTypeText}; and
	 * {@value org.tanjakostic.jcleancim.docgen.collector.WAX#A_type} and
	 * {@value org.tanjakostic.jcleancim.docgen.collector.WAX#A_typeKind} keys if enabled (
	 * <code>all=true</code>) and where they exist.
	 *
	 * @param mmType
	 * @param outRawData
	 *            in/out argument, filled here.
	 * @param showTorENUM
	 *            set this one to true when writing index for LNs and the attribute is a transient
	 *            CDC; or when printing actual CDC tables and the attribute is some kind of enum.
	 * @param trimPrimitivePrefix
	 *            set this one to true when dealing with an FCDA as type, for both CDC tables and
	 *            CDC attributes index - because we search for actual underlying types for e.g.
	 *            enums and composed DAs, for primitive/structured DAs it's the "real" primitive
	 *            type that gets picked (with {@value UML#PREF_P_}/{@value UML#PREF_S_}), and we
	 *            don't want to display that one in 7-3.
	 * @param all
	 *            if true, will set all potential fields; otherwise, will set only
	 *            {@value org.tanjakostic.jcleancim.docgen.collector.WAX#A_deducedTypeText}.
	 */
	protected void deduceTypeText(UmlClass mmType, RawData outRawData, boolean showTorENUM,
			boolean trimPrimitivePrefix, boolean all) {
		if (mmType.isCodedEnumFCDA() || mmType.isCodedEnumDA() || mmType.isCodedEnum()) {
			String enumTypeName = deduceUnderlyingTypeName(mmType, mmType.isCodedEnumFCDA(),
					trimPrimitivePrefix);
			if (all) {
				outRawData.putCell(WAX.A_type, enumTypeName);
				outRawData.putCell(WAX.A_typeKind, WAX.V_typeKind_CODED_ENUM);
			}
			String dtt = (showTorENUM) ? String.format("%s (%s)", WAX.V_typeKind_CODED_ENUM,
					enumTypeName) : enumTypeName;
			outRawData.putCell(WAX.A_deducedTypeText, dtt);
		} else if (mmType.isEnumFCDA() || mmType.isEnumDA() || mmType.isEnumeration()) {
			String enumTypeName = deduceUnderlyingTypeName(mmType, mmType.isEnumFCDA(),
					trimPrimitivePrefix);
			if (all) {
				outRawData.putCell(WAX.A_type, enumTypeName);
				outRawData.putCell(WAX.A_typeKind, WAX.V_typeKind_ENUM);
			}
			String dtt = (showTorENUM) ? String
					.format("%s (%s)", WAX.V_typeKind_ENUM, enumTypeName) : enumTypeName;
					outRawData.putCell(WAX.A_deducedTypeText, dtt);
		} else if (mmType.isPackedListFCDA() || mmType.isPackedListDA() || mmType.isPackedList()) {
			String packedListTypeName = deduceUnderlyingTypeName(mmType, mmType.isPackedListFCDA(),
					trimPrimitivePrefix);
			if (all) {
				outRawData.putCell(WAX.A_type, packedListTypeName);
				outRawData.putCell(WAX.A_typeKind, WAX.V_typeKind_PACKED_LIST);
			}
			outRawData.putCell(WAX.A_deducedTypeText, packedListTypeName);
		} else if (mmType.isComposedFCDA() || mmType.isComposedDA() || mmType.isStructured()) {
			String composedTypeName = deduceUnderlyingTypeName(mmType, mmType.isComposedFCDA(),
					trimPrimitivePrefix);
			if (all) {
				outRawData.putCell(WAX.A_type, composedTypeName);
				outRawData.putCell(WAX.A_typeKind, WAX.V_typeKind_CONSTRUCTED);
			}
			outRawData.putCell(WAX.A_deducedTypeText, composedTypeName);
		} else if (mmType.isAnyFCDA() || mmType.isPrimitiveDA() || mmType.isBasic()) {
			String typeName = deduceUnderlyingTypeName(mmType, mmType.isAnyFCDA(),
					trimPrimitivePrefix);
			if (all) {
				outRawData.putCell(WAX.A_type, typeName);
				outRawData.putCell(WAX.A_typeKind, WAX.V_typeKind_BASIC);
			}
			outRawData.putCell(WAX.A_deducedTypeText, typeName);
		} else if (mmType.isTransientCDC()) {
			String trans = (showTorENUM) ? (" (" + TRANSIENT + ")") : "";
			String typeName = stripTransientFromCdcName(mmType);
			if (all) {
				outRawData.putCell(WAX.A_type, typeName);
			}
			outRawData.putCell(WAX.A_deducedTypeText, typeName + trans);
		} else if (mmType.isEnumCDC() || mmType.isTrackingDerivedCDC()) {
			String derivedTypeName = splitDerivedCdcName(mmType);

			String superName = "";
			Collection<UmlClass> supers = mmType.getSuperclasses();
			if (!supers.isEmpty()) {
				superName = supers.iterator().next().getName();
			}
			if (all) {
				outRawData.putCell(WAX.A_type, superName);
				outRawData.putCell(WAX.A_underlyingType, derivedTypeName);
			}
			outRawData.putCell(WAX.A_deducedTypeText,
					String.format("%s (%s)", superName, derivedTypeName));
		} else {
			String typeName = deduceUnderlyingTypeName(mmType, false, trimPrimitivePrefix);
			if (all) {
				outRawData.putCell(WAX.A_type, typeName);
			}
			outRawData.putCell(WAX.A_deducedTypeText, typeName);
		}
	}

	/**
	 * "Descends" the UML type's private attributes to come to the actual primitive or enumerated
	 * type to display (those ending with "Kind" or starting with
	 * {@value org.tanjakostic.jcleancim.model.UML#PREF_P_} or
	 * {@value org.tanjakostic.jcleancim.model.UML#PREF_S_}).
	 *
	 * @param inType
	 *            type of attribute to analyse - could be (Packed)Enumeration, (Packed)EnumDA or
	 *            (Packed)EnumFCDA.
	 * @param isFcda
	 *            whether the passed type is an FCDA.
	 * @param trimPrimitivePrefix
	 *            whether to trim {@value org.tanjakostic.jcleancim.model.UML#PREF_P_} or
	 *            {@value org.tanjakostic.jcleancim.model.UML#PREF_S_}.
	 * @return formatted name with the actual (primitive) enumerated type to display.
	 */
	private static final String deduceUnderlyingTypeName(UmlClass inType, boolean isFcda,
			boolean trimPrimitivePrefix) {
		UmlClass type = inType;
		if (isFcda) {
			Set<UmlAttribute> attributes = type.findAttributes(UML.ATTR_attr);
			if (!attributes.isEmpty()) {
				UmlAttribute first = attributes.iterator().next();
				if (!first.isPublic()) {
					type = first.getType();
				}
			}
		}
		Set<UmlAttribute> attributes = type.findAttributes(UML.ATTR_val);
		if (!attributes.isEmpty()) {
			UmlAttribute first = attributes.iterator().next();
			if (!first.isPublic()) {
				type = first.getType();
			}
		}

		String result = type.getName();
		if (trimPrimitivePrefix) {
			result = trimPrimitivePrefix(result);
		}
		return result;
	}

	private static String trimPrimitivePrefix(String result) {
		if (result.startsWith(UML.PREF_P_)) {
			return result.replaceFirst(UML.PREF_P_, "");
		}
		if (result.startsWith(UML.PREF_S_)) {
			return result.replaceFirst(UML.PREF_S_, "");
		}
		return result;
	}

	/**
	 * Strips {@link UML#SUFF_Transient} from the class name, to "leave" only the CDC name, as
	 * required for writing in IEC61850-7-4.
	 */
	private static String stripTransientFromCdcName(UmlClass type) {
		String mmCdcName = type.getName();
		int idx = mmCdcName.lastIndexOf(UML.SUFF_Transient);
		String cdcName = mmCdcName;
		if (idx != -1) {
			cdcName = mmCdcName.substring(0, idx);
		}
		return cdcName;
	}

	/**
	 * Splits an enumerated CDC name to abstract CDC name (as required for writing in IEC 61850-7-4,
	 * and as defined in IEC61850-7-3) and (primitive) enumerated type.
	 * <p>
	 * Implementation note:
	 *
	 * @param inType
	 * @return formatted name with the actual (primitive) type to display.
	 */
	private static String splitDerivedCdcName(UmlClass inType) {
		int cdcNameLen = 3;
		UmlClass type = inType;

		UmlModel model = type.getModel();
		// we don't have attributes of the type derivedDA, so searching from name:
		String derivedDAName = type.getName().substring(cdcNameLen);
		Set<UmlClass> derivedDAs = model.findClasses(derivedDAName);
		if (!derivedDAs.isEmpty()) {
			type = derivedDAs.iterator().next();
		}
		return deduceUnderlyingTypeName(type, false, true);
	}

	// ============= presence conditions ======================

	/**
	 * Fills appropriately raw data 'presCond', 'presCondArgs', 'presCondArgsID' and 'cond'.
	 */
	protected static void fillPresenceConditionAndArgs(RawData entry, PresenceCondition pc,
			boolean isInherited, String context) {
		fillPresenceConditionAndArgs(entry, pc, isInherited, context, false);
	}

	/**
	 * Fills appropriately raw data 'presCond', 'presCondArgs', 'presCondArgsID' and 'cond' if
	 * <code>isDerivedStats=false</code>, otherwise 'dsPresCond', 'dsPresCondArgs',
	 * 'dsPresCondArgsID' and 'dsCond'.
	 */
	protected static void fillPresenceConditionAndArgs(RawData entry, PresenceCondition pc,
			boolean isInherited, String context, boolean isDerivedStats) {
		String pcTag = isDerivedStats ? "dsPresCond" : WAX.A_presCond;
		String pcArgTag = isDerivedStats ? "dsPresCondArgs" : "presCondArgs";
		String condTag = isDerivedStats ? "dsCond" : "cond";
		String pcArgIDTag = isDerivedStats ? "dsPresCondArgsID" : "presCondArgsID";

		entry.putCellNonEmpty(pcTag, pc.getStem());
		entry.putCellNonEmpty(pcArgTag, pc.getArgs());
		String cond = pc.getText();
		if (Util.hasContent(cond)) {
			if (!isInherited) {
				entry.putCellNonEmpty(condTag, cond);
				String docId = (pc.getConstraint() == null) ? "" : DocIDHelper.createDocId(
						pc.getConstraint(), condTag + "." + context);
				entry.putCellNonEmpty(pcArgIDTag, docId);
			} else {
				entry.putCellNonEmpty(condTag, "");
			}
		}
	}
}
