/**
 * Copyright (C) 2009-2015 Tatjana (Tanja) Kostic
 * <p>
 * This file belongs to jCleanCim, a tool supporting tasks of UML model managers for IEC TC57 CIM
 * and 61850 models.
 * <p>
 * This program is free software: you can redistribute it and/or modify it under the terms of the
 * GNU Lesser General Public License version 3.
 * <p>
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details. You should have received a copy of the GNU Lesser
 * General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

package org.tanjakostic.jcleancim.docgen.writer;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;

/**
 * Templates for doc generation have to use labels to indicate where to insert the documentation of
 * what element of the UML model into the output document. Currently needed and recognised labels to
 * be used in the input templates are in the following format:
 *
 * <pre>
 *   startUmlDiagram.{packageName}.{diagramName}.endUml
 *   startUmlAttribute.{className}.{attributeName}.endUml
 *   startUmlFile..endUml
 *
 *   startUmlPresenceConditions.{packageName}.endUml (IEC 61850-7-3, for presence conditions table)
 *   startUmlFCs.{packageName}.endUml                (IEC 61850-7-3 and IEC 61850-7-2, for FC table)
 *   startUmlTrgOps.{packageName}.endUml             (IEC 61850-7-2, for TrgOp table)
 *   startUmlAbbreviations.{packageName}.endUml      (IEC 61850-7-4, for DO abbreviations table)
 *   startUmlSclEnums.{packageName}.endUml           (IEC 61850-7-4, 7-3, for enums as XML)
 *
 *   startUmlPackage.{packageName}.endUml
 *   startUmlDataIndex.{packageName}.endUml          (for IEC 61850-7-4,3, for data semantics tables)
 *   startUmlLNMapPackage.{packageName}.endUml       (for IEC 61850-7-4, for function/LN map tables)
 * </pre>
 *
 * The tokens enlosed in curly braces are the names of UML elements designating what needs to be
 * inserted in place of the whole above string.
 * <p>
 * This format avoids us the need to define bookmarks in the input document (tedious and
 * error-prone) and makes it simple to sequentially search the input document and insert the text
 * and diagrams as they come.
 * <p>
 * <b>Usage</b>
 * <p>
 * Instance of kind {@link Kind#UNSUPPORTED} always returns non-null error that you may want to use
 * to replace the placeholder to indicate failure. Instances of kind {@link Kind#FILE} are the
 * simplest as they need not parse anything, so no errors and no saved tokens.
 * <p>
 * The instances of other kinds do need to parse the placeholder and will have tokens set as
 * follows:
 * <ul>
 * <li>first token ({@link Kind#PRES_CONDITIONS}, {@link Kind#ABBREVIATIONS}, {@link Kind#SCL_ENUMS}, {@link Kind#PACKAGE}, {@link Kind#LNMAP_PACKAGE}, {@link Kind#DATA_INDEX}, {@link Kind#FCS},
 * {@link Kind#TRGOPS}) or
 * <li>both tokens ({@link Kind#ATTRIBUTE}, {@link Kind#DIAGRAM}).
 * </ul>
 * In case the parsed token(s) is null or empty, the instance will contain a non-null error string.
 * So, you will want to check first for error (null means no errors) before passing tokens for
 * search, and in case of error, you may want to replace the placeholder to indicate failure.
 * <p>
 * <b>Important:</b> To have correct headings and paragraph formats, ensure you use the following
 * placeholders in a heading paragraph:
 * <ul>
 * <li>{@link Kind#PACKAGE},
 * <li>{@link Kind#PRES_CONDITIONS},
 * <li>{@link Kind#FCS},
 * <li>{@link Kind#TRGOPS},
 * <li>{@link Kind#SCL_ENUMS},
 * <li>{@link Kind#LNMAP_PACKAGE}, and,
 * <li>{@link Kind#DATA_INDEX}.
 * </ul>
 *
 * @author tatjana.kostic@ieee.org
 * @version $Id: Placeholder.java 2728 2015-05-14 20:00:03Z CHTAKOS $
 */
public class Placeholder {

	public enum Kind {
		FILE,
		DIAGRAM,
		ATTRIBUTE,
		PRES_CONDITIONS,
		FCS,
		TRGOPS,
		ABBREVIATIONS,
		SCL_ENUMS,
		PACKAGE,
		DATA_INDEX,
		LNMAP_PACKAGE,
		UNSUPPORTED;

		/** Returns whether <code>kind</code> is expected to be in the heading paragraph. */
		public static boolean isForHeading(Kind kind) {
			return EnumSet.of(PACKAGE, PRES_CONDITIONS, TRGOPS, LNMAP_PACKAGE, DATA_INDEX)
					.contains(kind);
		}
	}

	private static final String START_UML = "startUml";
	private static final String END_UML = "endUml";
	private static final String SEPARATOR = ".";

	public static final String PH_LETTERS = "[ADFLTPS]";
	public static final String MS_PATTERN = String.format("%s%s*%s*%s%s", START_UML, PH_LETTERS,
			SEPARATOR, SEPARATOR, END_UML);

	private static final String UML_FILE = START_UML + "File";
	private static final String UML_DIAGRAM = START_UML + "Diagram";
	private static final String UML_ATTRIBUTE = START_UML + "Attribute";
	private static final String UML_PRES_CONDITIONS = START_UML + "PresenceConditions";
	private static final String UML_FCS = START_UML + "FCs";
	private static final String UML_TRGOPS = START_UML + "TrgOps";
	private static final String UML_ABBREVIATIONS = START_UML + "Abbreviations";
	private static final String UML_SCL_ENUMS = START_UML + "SclEnums";
	private static final String UML_PACKAGE = START_UML + "Package";
	private static final String UML_LNMAP_PACKAGE = START_UML + "LNMapPackage";
	private static final String UML_DATA_INDEX = START_UML + "DataIndex";

	static String constructFilePlaceholderText() {
		return UML_FILE + SEPARATOR + SEPARATOR + END_UML;
	}

	static String constructDiagramPlaceholderText(String pckName, String diagName) {
		return UML_DIAGRAM + SEPARATOR + pckName + SEPARATOR + diagName + SEPARATOR + END_UML;
	}

	static String constructAttributePlaceholderText(String className, String attrName) {
		return UML_ATTRIBUTE + SEPARATOR + className + SEPARATOR + attrName + SEPARATOR + END_UML;
	}

	static String constructPresConditionsPackagePlaceholderText(String pckName) {
		return UML_PRES_CONDITIONS + SEPARATOR + pckName + SEPARATOR + END_UML;
	}

	static String constructFcsPackagePlaceholderText(String pckName) {
		return UML_FCS + SEPARATOR + pckName + SEPARATOR + END_UML;
	}

	static String constructTrgOpsPackagePlaceholderText(String pckName) {
		return UML_TRGOPS + SEPARATOR + pckName + SEPARATOR + END_UML;
	}

	static String constructAbbrPackagePlaceholderText(String pckName) {
		return UML_ABBREVIATIONS + SEPARATOR + pckName + SEPARATOR + END_UML;
	}

	static String constructEnumPackagePlaceholderText(String pckName) {
		return UML_SCL_ENUMS + SEPARATOR + pckName + SEPARATOR + END_UML;
	}

	static String constructPackagePlaceholderText(String pckName) {
		return UML_PACKAGE + SEPARATOR + pckName + SEPARATOR + END_UML;
	}

	static String constructLNMapPackagePlaceholderText(String pckName) {
		return UML_LNMAP_PACKAGE + SEPARATOR + pckName + SEPARATOR + END_UML;
	}

	static String constructDataIndexPlaceholderText(String pckName) {
		return UML_DATA_INDEX + SEPARATOR + pckName + SEPARATOR + END_UML;
	}

	public static List<String> getSupportedFormats() {
		List<String> result = new ArrayList<String>();
		result.add(constructFilePlaceholderText());
		result.add(constructAttributePlaceholderText("className", "attrName"));
		result.add(constructDiagramPlaceholderText("pckName", "diagName"));
		result.add(constructPresConditionsPackagePlaceholderText("pckName"));
		result.add(constructFcsPackagePlaceholderText("pckName"));
		result.add(constructTrgOpsPackagePlaceholderText("pckName"));
		result.add(constructAbbrPackagePlaceholderText("pckName"));
		result.add(constructEnumPackagePlaceholderText("pckName"));
		result.add(constructPackagePlaceholderText("pckName"));
		result.add(constructLNMapPackagePlaceholderText("pckName"));
		result.add(constructDataIndexPlaceholderText("pckName"));
		return result;
	}

	private static final String TOKEN_ERROR_TEXT = "Null or empty token";
	private static final String PLACEHOLDER_ERROR_TEXT = "Unrecognised placeholder";
	private static final String NOT_IN_MODEL_ERROR_TEXT = " not found in model";

	private final String _text;
	private Kind _kind;
	private String _firstToken;
	private String _secondToken;
	private String _errorText;
	private String _replacedText;
	private int _figureCountBefore;
	private int _tableCountBefore;
	private int _myFigureCount;
	private int _myTableCount;

	public Placeholder(String text) {
		this(text, 0, 0);
	}

	public Placeholder(String text, int figureCountBefore, int tableCountBefore) {
		_text = text;
		_figureCountBefore = figureCountBefore;
		_tableCountBefore = tableCountBefore;
		_myFigureCount = 0;
		_myTableCount = 0;
		if (text.startsWith(Placeholder.UML_FILE)) {
			init(Kind.FILE, null, null, null, null);
		} else if (text.startsWith(Placeholder.UML_ATTRIBUTE)) {
			String[] body = parseNames();
			String className = (body != null && body.length == 2) ? body[0] : null;
			String attrName = (body != null && body.length == 2) ? body[1] : null;
			String error = (isNullOrEmpty(className) || isNullOrEmpty(attrName)) ? formatErrMessage()
					: null;
			init(Kind.ATTRIBUTE, className, attrName, error, null);
		} else if (text.startsWith(Placeholder.UML_DIAGRAM)) {
			String[] body = parseNames();
			String packageName = (body != null && body.length == 2) ? body[0] : null;
			String diagramName = (body != null && body.length == 2) ? body[1] : null;
			String error = (isNullOrEmpty(packageName) || isNullOrEmpty(diagramName)) ? formatErrMessage()
					: null;
			init(Kind.DIAGRAM, packageName, diagramName, error, null);
		} else if (text.startsWith(Placeholder.UML_PRES_CONDITIONS)) {
			String[] body = parseNames();
			String packageName = (body != null && body.length == 1) ? body[0] : null;
			String error = isNullOrEmpty(packageName) ? formatErrMessage() : null;
			init(Kind.PRES_CONDITIONS, packageName, null, error, null);
		} else if (text.startsWith(Placeholder.UML_FCS)) {
			String[] body = parseNames();
			String packageName = (body != null && body.length == 1) ? body[0] : null;
			String error = isNullOrEmpty(packageName) ? formatErrMessage() : null;
			init(Kind.FCS, packageName, null, error, null);
		} else if (text.startsWith(Placeholder.UML_TRGOPS)) {
			String[] body = parseNames();
			String packageName = (body != null && body.length == 1) ? body[0] : null;
			String error = isNullOrEmpty(packageName) ? formatErrMessage() : null;
			init(Kind.TRGOPS, packageName, null, error, null);
		} else if (text.startsWith(Placeholder.UML_ABBREVIATIONS)) {
			String[] body = parseNames();
			String packageName = (body != null && body.length == 1) ? body[0] : null;
			String error = isNullOrEmpty(packageName) ? formatErrMessage() : null;
			init(Kind.ABBREVIATIONS, packageName, null, error, null);
		} else if (text.startsWith(Placeholder.UML_SCL_ENUMS)) {
			String[] body = parseNames();
			String packageName = (body != null && body.length == 1) ? body[0] : null;
			String error = isNullOrEmpty(packageName) ? formatErrMessage() : null;
			init(Kind.SCL_ENUMS, packageName, null, error, null);
		} else if (text.startsWith(Placeholder.UML_PACKAGE)) {
			String[] body = parseNames();
			String packageName = (body != null && body.length == 1) ? body[0] : null;
			String error = isNullOrEmpty(packageName) ? formatErrMessage() : null;
			init(Kind.PACKAGE, packageName, null, error, null);
		} else if (text.startsWith(Placeholder.UML_LNMAP_PACKAGE)) {
			String[] body = parseNames();
			String packageName = (body != null && body.length == 1) ? body[0] : null;
			String error = isNullOrEmpty(packageName) ? formatErrMessage() : null;
			init(Kind.LNMAP_PACKAGE, packageName, null, error, null);
		} else if (text.startsWith(Placeholder.UML_DATA_INDEX)) {
			String[] body = parseNames();
			String packageName = (body != null && body.length == 1) ? body[0] : null;
			String error = isNullOrEmpty(packageName) ? formatErrMessage() : null;
			init(Kind.DATA_INDEX, packageName, null, error, null);
		} else {
			init(Kind.UNSUPPORTED, null, null, formatUnsupportedMessage(), null);
		}
	}

	private boolean isNullOrEmpty(String token) {
		return token == null || "null".equals(token) || token.isEmpty();
	}

	private void init(Kind kind, String name1, String name2, String error, String replacedText) {
		_kind = kind;
		_firstToken = name1;
		_secondToken = name2;
		_errorText = error;
		_replacedText = replacedText;
	}

	private String formatErrMessage() {
		return formatErrMessage(TOKEN_ERROR_TEXT);
	}

	private String formatUnsupportedMessage() {
		return formatErrMessage(PLACEHOLDER_ERROR_TEXT);
	}

	private String formatErrMessage(String errorText) {
		return "$ERROR " + _text + ": " + errorText + "$";
	}

	private String formatModelErrMessage() {
		return "$ERROR " + toString() + NOT_IN_MODEL_ERROR_TEXT + "$";
	}

	private String[] parseNames() {
		String[] body = _text.split("\\" + SEPARATOR);
		List<String> result = new ArrayList<String>();
		for (int i = 1; i < body.length - 1; ++i) {
			result.add(body[i]);
		}
		return result.toArray(new String[0]);
	}

	// ----------------- API -----------------

	/**
	 * Returns the placeholder text.
	 */
	public String getText() {
		return _text;
	}

	public Kind getKind() {
		return _kind;
	}

	public String getFirstToken() {
		return _firstToken;
	}

	public String getSecondToken() {
		return _secondToken;
	}

	/**
	 * Returns null if there are no parsing errors, the error message otherwise.
	 */
	public String getErrorText() {
		return _errorText;
	}

	/**
	 * Use this setter when finder could not find valid tokens in the model (e.g., format of the
	 * placeholder is ok, but the names do not match elements in the model).
	 */
	public void updateModelErrorText() {
		_errorText = formatModelErrMessage();
	}

	/**
	 * Use this setter to track progress and status of placeholder replacement.
	 */
	public void setReplacedText(String replacedText) {
		_replacedText = replacedText;
	}

	/**
	 * Returns number of figures (i.e., found captions starting with "Figure") before this
	 * placeholder.
	 */
	public int getFigureCountBefore() {
		return _figureCountBefore;
	}

	/**
	 * Returns number of tables (i.e., found captions starting with "Table") before this
	 * placeholder.
	 */
	public int getTableCountBefore() {
		return _tableCountBefore;
	}

	/**
	 * Returns the replacement text, as set by the user with {@link #setReplacedText(String)}.
	 * Initial value is null and is never changed by this class itself.
	 */
	public String getReplacedText() {
		return _replacedText;
	}

	/**
	 * Increments the number of figures before this placeholder.
	 */
	public void incrementFigureBefore() {
		++_figureCountBefore;
	}

	/**
	 * Increments the number of tables before this placeholder.
	 */
	public void incrementTableBefore() {
		++_tableCountBefore;
	}

	/**
	 * Returns the index to be used to reference the added figure caption.
	 */
	public int addFigure() {
		++_myFigureCount;
		return getFigureCount();
	}

	/**
	 * Returns the last figure caption index by the current end of the range.
	 */
	public int getFigureCount() {
		return getFigureCountBefore() + _myFigureCount;
	}

	/**
	 * Returns the index to be used to reference the added table caption.
	 */
	public int addTable() {
		++_myTableCount;
		return getTableCount();
	}

	/**
	 * Returns the last table caption index by the current end of the range.
	 */
	public int getTableCount() {
		return getTableCountBefore() + _myTableCount;
	}

	@Override
	public String toString() {
		if (_errorText != null) {
			return _errorText;
		}
		StringBuilder sb = new StringBuilder(_kind.toString());
		if (_firstToken != null) {
			sb.append(" ").append(_firstToken);
			if (_secondToken != null) {
				sb.append(SEPARATOR).append(_secondToken);
			}
		}
		if (_figureCountBefore != 0 || _myFigureCount != 0) {
			sb.append(", figures (");
			if (_figureCountBefore != 0) {
				sb.append(_figureCountBefore + " before ");
			}
			if (_myFigureCount != 0) {
				sb.append(_myFigureCount + " mine");
			}
			sb.append(")");
		}
		if (_tableCountBefore != 0 || _myTableCount != 0) {
			sb.append(", tables (");
			if (_tableCountBefore != 0) {
				sb.append(_tableCountBefore + " before ");
			}
			if (_myTableCount != 0) {
				sb.append(_myTableCount + " mine");
			}
			sb.append(")");
		}
		return sb.toString();
	}
}
