/**
 * Copyright (C) 2009-2015 Tatjana (Tanja) Kostic
 * <p>
 * This file belongs to jCleanCim, a tool supporting tasks of UML model managers for IEC TC57 CIM
 * and 61850 models.
 * <p>
 * This program is free software: you can redistribute it and/or modify it under the terms of the
 * GNU Lesser General Public License version 3.
 * <p>
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details. You should have received a copy of the GNU Lesser
 * General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

package org.tanjakostic.jcleancim.docgen.writer;

import java.util.Arrays;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

/**
 * Maps document in-built styles as replacement for our desired styles (in IEC template), to allow
 * for doc generation even with a non-IEC template.
 * <p>
 * FIXME: Application needs to initialise the name ({@link #initialiseFromExisting(Map)} so that
 * {@link #getName()} returns desired style name if that style is available in the document,
 * otherwise the default style name.
 *
 * @author tatjana.kostic@ieee.org
 * @version $Id: Style.java 2728 2015-05-14 20:00:03Z CHTAKOS $
 */
public enum Style {

	para("PARAGRAPH", "Normal"),
	fig("Picture", "Normal"),
	tabhead("TABLE-col-heading", "Normal"),
	tabcell("TABLE-cell", "Normal"),
	figcapt("FIGURE-title", "Caption"),
	tabcapt("TABLE-title", "Caption"),
	h1("Heading 1", "Heading 1"),
	h2("Heading 2", "Heading 2"),
	h3("Heading 3", "Heading 3"),
	h4("Heading 4", "Heading 4"),
	h5("Heading 5", "Heading 5"),
	h6("Heading 6", "Heading 6"),
	h7("Heading 7", "Heading 7"),
	h8("Heading 8", "Heading 8"),
	h9("Heading 9", "Heading 9");

	private static final Logger _logger = Logger.getLogger(Style.class.getName());

	public static final String TOC_PREFIX = "TOC";

	/** Initialises predefined enum literals with the actual data (e.g., read from an opened file). */
	public static void initialiseFromExisting(Map<String, String> idPerName) {
		for (Style style : Style.values()) {
			String wantedName = style.getWantedName();
			if (idPerName.containsKey(wantedName)) {
				style._wantedStyleId = idPerName.get(wantedName);
				style._wantedFound = true;
				continue;
			}

			/** In .docx formats, names of some styles are stored in lower case... */
			wantedName = wantedName.toLowerCase();
			if (idPerName.containsKey(wantedName)) {
				style._wantedName = wantedName;
				style._wantedStyleId = idPerName.get(wantedName);
				style._wantedFound = true;
				continue;
			}

			style._wantedFound = false;
			_logger.info("wanted style name not found: " + style.getWantedName()
					+ ". Will be using a builtin style instead.");
			String builtinName = style.getBuiltinName();
			if (idPerName.containsKey(builtinName)) {
				style._builtinStyleId = idPerName.get(builtinName);
				continue;
			}

			builtinName = builtinName.toLowerCase();
			if (idPerName.containsKey(builtinName)) {
				style._builtinName = builtinName;
				style._builtinStyleId = idPerName.get(builtinName);
				continue;
			}
			_logger.error("failed to find the builtin style " + style.getWantedName()
					+ " in the document. Doc generation may fail ...");
		}
	}

	public static boolean isTOC(String styleName) {
		return styleName.toLowerCase().startsWith(TOC_PREFIX.toLowerCase());
	}

	private String _wantedName;
	private String _wantedStyleId;
	private String _builtinName;
	private String _builtinStyleId;
	private boolean _wantedFound;

	Style(String wantedName, String builtinName) {
		_wantedName = wantedName;
		_wantedStyleId = null;
		_builtinName = builtinName;
		_builtinStyleId = null;
	}

	public static Style getHeadingStyle(int outlineLevel) {
		switch (outlineLevel) {
			case 1:
				return h1;
			case 2:
				return h2;
			case 3:
				return h3;
			case 4:
				return h4;
			case 5:
				return h5;
			case 6:
				return h6;
			case 7:
				return h7;
			case 8:
				return h8;
			case 9:
				return h9;
			default:
				return null;
		}
	}

	public static boolean usableForFigureCaption(String styleNameOrId) {
		return usableForCaption(styleNameOrId, figcapt);
	}

	public static boolean usableForTableCaption(String styleNameOrId) {
		return usableForCaption(styleNameOrId, tabcapt);
	}

	private static boolean usableForCaption(String styleNameOrId, Style style) {
		boolean foundByName = style.getAllKnownNames().contains(styleNameOrId);
		if (foundByName) {
			return true;
		}
		return style.getAllKnownIds().contains(styleNameOrId);
	}

	// ---------------------------------

	/**  */
	public String getWantedName() {
		return _wantedName;
	}

	/** For .docx only. */
	public String getWantedStyleId() {
		return _wantedStyleId;
	}

	public String getBuiltinName() {
		return _builtinName;
	}

	/** For .docx only. */
	public String getBuiltinStyleId() {
		return _builtinStyleId;
	}

	public List<String> getAllKnownNames() {
		return Arrays.asList(getBuiltinName(), getWantedName());
	}

	public List<String> getAllKnownIds() {
		return Arrays.asList(getBuiltinStyleId(), getWantedStyleId());
	}

	/** Returns usable style name (both .doc and .docs). */
	public String getName() {
		return _wantedFound ? getWantedName() : getBuiltinName();
	}

	/** Returns style ID (only .docs). */
	public String getStyleId() {
		return _wantedFound ? getWantedStyleId() : getBuiltinStyleId();
	}
}
