/**
 * Copyright (C) 2009-2015 Tatjana (Tanja) Kostic
 * <p>
 * This file belongs to jCleanCim, a tool supporting tasks of UML model managers for IEC TC57 CIM
 * and 61850 models.
 * <p>
 * This program is free software: you can redistribute it and/or modify it under the terms of the
 * GNU Lesser General Public License version 3.
 * <p>
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details. You should have received a copy of the GNU Lesser
 * General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

package org.tanjakostic.jcleancim.docgen.writer.word;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.tanjakostic.jcleancim.common.Config;
import org.tanjakostic.jcleancim.docgen.UnsupportedInputFormatException;
import org.tanjakostic.jcleancim.docgen.UnsupportedOutputFormatException;
import org.tanjakostic.jcleancim.docgen.collector.ClassDoc;
import org.tanjakostic.jcleancim.docgen.collector.FigureDoc;
import org.tanjakostic.jcleancim.docgen.collector.PackageDoc;
import org.tanjakostic.jcleancim.docgen.collector.PackageScl;
import org.tanjakostic.jcleancim.docgen.collector.PropertiesDoc;
import org.tanjakostic.jcleancim.docgen.writer.AbstractWriter;
import org.tanjakostic.jcleancim.docgen.writer.Caption;
import org.tanjakostic.jcleancim.docgen.writer.Cursor;
import org.tanjakostic.jcleancim.docgen.writer.CursorList;
import org.tanjakostic.jcleancim.docgen.writer.Placeholder;
import org.tanjakostic.jcleancim.docgen.writer.Range;
import org.tanjakostic.jcleancim.docgen.writer.Style;
import org.tanjakostic.jcleancim.docgen.writer.Caption.CaptionKind;
import org.tanjakostic.jcleancim.model.TextDescription;
import org.tanjakostic.jcleancim.model.TextDescription.TextKind;
import org.tanjakostic.jcleancim.util.Util;

/**
 * @param <O>
 *            technology-specific type to access range object.
 * @author tatjana.kostic@ieee.org
 * @version $Id: AbstractWordWriter.java 2729 2015-05-14 20:08:13Z CHTAKOS $
 */
abstract public class AbstractWordWriter<O> extends AbstractWriter implements WordWriter<O>,
WordHelper<O> {
	private static final Logger _logger = Logger.getLogger(AbstractWordWriter.class.getName());

	private final WordWriterInput _input;

	/**
	 * Constructor; copies input template into output file that will be filled with this writer.
	 *
	 * @param input
	 */
	protected AbstractWordWriter(WordWriterInput input) throws UnsupportedInputFormatException,
	UnsupportedOutputFormatException, IOException {
		super(input);
		_input = input;

		File src = new File(getInput().getInTemplatePath());
		File dst = new File(getInput().getOutFilePath());

		if (!getSupportedFormats().contains(Util.getFileExtensionWithDot(src.getName()))) {
			throw new UnsupportedInputFormatException("Supporting " + getSupportedFormats()
					+ " as template, received " + src.getName() + ".");
		}
		if (!getSupportedFormats().contains(Util.getFileExtensionWithDot(dst.getName()))) {
			throw new UnsupportedOutputFormatException("Supporting " + getSupportedFormats()
					+ " as output, required to process " + dst.getName() + ".");
		}

		copyTemplateAsOutput(src, dst);
		_logger.info("Created " + getClass().getSimpleName() + ".");
	}

	private void copyTemplateAsOutput(File src, File dst) throws IOException {
		Util.copy(src, dst);
	}

	private void doIt(boolean scanOnly, PostProcessor pp) {
		try {
			createWordApp();
			openDoc();
			_logger.info(getWordAppName() + " version: " + getWordAppVersion());
			_logger.info("");

			initDocgenOptimisationOptions();
			setCustomDocProperties(getDocumentMetadata());
			initStyleNames();

			// When using the close-reopen optimisation, we MUST NOT do field updates immediately
			// after opening it; if we do, in CIM Word templates we have the last heading
			// (Bibliography) that gets wrongly picked immediatly after the first close/reopen. In
			// other templates, similar happens, because Word will internally increase range for the
			// TOC in the document before saving it, but after reopening, the doc is shorter ! We
			// have no programmatical means to track this condition.
			//
			// Because it seems that we always want to use close-reopen optimisation with binary
			// docs, and with XML docs we don't do field update anyway, I'm commenting this piece
			// of code (but I keep it in case we need to test something):
			//
			// if (!applyCloseReopen()) {
			// updateAllFields();
			// }

			setDocgenOptimisationOptions();

			List<Range<O>> figCaptionRanges = new ArrayList<Range<O>>();
			List<Range<O>> tabCaptionRanges = new ArrayList<Range<O>>();
			collectFigureAndTableCaptionRanges(figCaptionRanges, tabCaptionRanges);
			_cursors = scanPlaceholderRanges(figCaptionRanges, tabCaptionRanges);

			if (!scanOnly) {
				insertDocIntoPlaceholders();
			}
			logReplacementFailures();

			if (pp != null) {
				pp.postProcess();
			}

			unsetDocgenOptimisationOptions();
			updateAllFields(getInput().isSkipTiming());
		} catch (Exception e) {
			_logger.error("##### Prematurely closing document" + " (and exiting MS Word)"
					+ " due to exception: " + e.getMessage());
			e.printStackTrace(System.err);
		} finally {
			try {
				closeDoc();
				exitAppAndSaveDocument();
			} catch (IOException e) {
				_logger.warn(e.getMessage());
			}
		}
	}

	private void logReplacementFailures() {
		List<String> replacementFailures = getCursors().getReplacementFailures();
		if (replacementFailures.isEmpty()) {
			return;
		}

		_logger.error(String.format("There were %d placeholder replacement failures:",
				Integer.valueOf(replacementFailures.size())));
		for (String failure : replacementFailures) {
			_logger.error(failure);
		}
		_logger.info("Supported placeholder formats:");
		_logger.info(Util.concatCharSeparatedTokens(Util.NL, false,
				Placeholder.getSupportedFormats()));
	}

	// --------- Word app and document lifecycle -------------

	/**
	 * {@inheritDoc}
	 * <p>
	 * This default implementation does nothing.
	 */
	@Override
	public void createWordApp() {
		// no-op
	}

	/**
	 * {@inheritDoc}
	 * <p>
	 * This default implementation does nothing.
	 */
	@Override
	public void exitAppAndSaveDocument() {
		// no-op
	}

	// ------------------- application & document optimisation options ---------------

	/**
	 * {@inheritDoc}
	 * <p>
	 * This default implementation does nothing.
	 */
	@Override
	public void initDocgenOptimisationOptions() {
		// no-op
	}

	/**
	 * {@inheritDoc}
	 * <p>
	 * This default implementation does nothing.
	 */
	@Override
	public void setDocgenOptimisationOptions() {
		// no-op
	}

	/**
	 * {@inheritDoc}
	 * <p>
	 * This default implementation does nothing.
	 */
	@Override
	public void unsetDocgenOptimisationOptions() {
		// no-op
	}

	// -------------------

	private void initStyleNames() {
		Map<String, String> idsPerName = getExistingStyleNames();
		_logger.debug("======== in initStyleNames() - existing styles {name, id}: ======");
		for (Entry<String, String> e : idsPerName.entrySet()) {
			_logger.debug(e.toString());
		}
		_logger.debug("DBG ==============");

		Style.initialiseFromExisting(idsPerName);
	}

	private void updateAllFields(boolean skipTiming) {
		_logger.info("   updating document fields (takes a while)...");
		long start = System.currentTimeMillis();
		updateFields();
		updateTablesOf("TablesOfContents");
		updateTablesOf("TablesOfFigures");
		Util.logCompletion(Level.INFO, "   updated document fields", start, skipTiming);
	}

	// ----------------- main processing steps -----------------

	/**
	 * Gets the map from Word and initialises caption range lists - homegrown range housekeeping for
	 * performance reasons.
	 */
	private void collectFigureAndTableCaptionRanges(List<Range<O>> figCaptionRanges,
			List<Range<O>> tabCaptionRanges) {
		Util.logSubtitle(Level.INFO, "collecting existing figure/table captions from template...");
		long start = System.currentTimeMillis();

		Map<CaptionKind, List<Caption<O>>> captions = collectCaptions();

		for (Caption<O> cd : captions.get(CaptionKind.Figure)) {
			figCaptionRanges.add(createRange(cd.getRange().getObject()));
		}
		for (Caption<O> cd : captions.get(CaptionKind.Table)) {
			tabCaptionRanges.add(createRange(cd.getRange().getObject()));
		}

		Util.logCompletion(Level.INFO, "collected existing figure/table captions from template.",
				start, getInput().isSkipTiming());
	}

	/**
	 * This one scans the whole initial document and initialises placeholders, with text, ranges and
	 * counts of existing tables and figures (captions) before each of them. This is essential for
	 * correct references to table and figure captions that we create on the fly.
	 */
	private CursorList<O> scanPlaceholderRanges(List<Range<O>> figCaptionRanges,
			List<Range<O>> tabCaptionRanges) {
		Util.logSubtitle(Level.INFO, "scanning placeholders...");
		long start = System.currentTimeMillis();

		CursorList<O> result = initCursorsFromPlaceholders(figCaptionRanges, tabCaptionRanges);

		Util.logCompletion(Level.INFO, "scanned placeholders.", start, getInput().isSkipTiming());
		return result;
	}

	/**
	 * When actually writing packages, we cath the exceptions and try to continue, so you see the
	 * most of result possible.
	 */
	private void insertDocIntoPlaceholders() {
		Util.logSubtitle(Level.INFO, "inserting documentation into placeholders");

		boolean isDeep = getInput().isDeep();
		for (int i = 0; i < getCursors().size(); ++i) {
			Cursor<O> cursor = getCursors().get(i);

			long start = System.currentTimeMillis();
			_logger.info(String.format("  replacing %s...", cursor));

			switch (cursor.getPlaceholder().getKind()) {
				case FILE: {
					String newText = getInput().getModelFileName();
					cursor.getRange().setText(newText);
					newText = cursor.getRange().getText();
					cursor.getPlaceholder().setReplacedText(newText);
					break;
				}
				case ATTRIBUTE: {
					String newText = determineTextForAttributePh(cursor.getPlaceholder());
					cursor.getRange().setText(newText);
					newText = cursor.getRange().getText();
					cursor.getPlaceholder().setReplacedText(newText);
					break;
				}
				case DIAGRAM: {
					String newText = determineTextForDiagramPhAndInsertDiagram(cursor.getRange(),
							cursor.getPlaceholder());
					if (newText.isEmpty()) {
						prependText(cursor.getRange(), newText);
					} else {
						cursor.getRange().setText(newText);
					}
					newText = cursor.getRange().getText();
					cursor.getPlaceholder().setReplacedText(newText);
					break;
				}
				case PRES_CONDITIONS: {
					String newText = determineTextForPackagePh(cursor.getPlaceholder());
					cursor.getRange().setText(newText);
					newText = cursor.getRange().getText();
					cursor.getPlaceholder().setReplacedText(newText);
					if (isDeep && cursor.getPlaceholder().getErrorText() == null) {
						try {
							cursor = writePresCondTable(cursor,
									getInput().getPackageDocs().get(newText));
						} catch (Exception e) {
							_logger.warn("#######  Caught exception - skipping: " + e.getMessage());
							e.printStackTrace(System.err);
						}
					}
					break;
				}
				case FCS: {
					String newText = determineTextForPackagePh(cursor.getPlaceholder());
					cursor.getRange().setText(newText);
					newText = cursor.getRange().getText();
					cursor.getPlaceholder().setReplacedText(newText);
					if (isDeep && cursor.getPlaceholder().getErrorText() == null) {
						try {
							cursor = writeFcTable(cursor, getInput().getPackageDocs().get(newText));
						} catch (Exception e) {
							_logger.warn("#######  Caught exception - skipping: " + e.getMessage());
							e.printStackTrace(System.err);
						}
					}
					break;
				}
				case TRGOPS: {
					String newText = determineTextForPackagePh(cursor.getPlaceholder());
					cursor.getRange().setText(newText);
					newText = cursor.getRange().getText();
					cursor.getPlaceholder().setReplacedText(newText);
					if (isDeep && cursor.getPlaceholder().getErrorText() == null) {
						try {
							cursor = writeTrgOpTable(cursor,
									getInput().getPackageDocs().get(newText));
						} catch (Exception e) {
							_logger.warn("#######  Caught exception - skipping: " + e.getMessage());
							e.printStackTrace(System.err);
						}
					}
					break;
				}
				case ABBREVIATIONS: {
					String newText = determineTextForPackagePh(cursor.getPlaceholder());
					cursor.getRange().setText("");
					String emptyText = cursor.getRange().getText();
					cursor.getPlaceholder().setReplacedText(emptyText);
					if (isDeep && cursor.getPlaceholder().getErrorText() == null) {
						try {
							cursor = writeAbbrTable(cursor, getInput().getPackageDocs()
									.get(newText));
						} catch (Exception e) {
							_logger.warn("#######  Caught exception - skipping: " + e.getMessage());
							e.printStackTrace(System.err);
						}
					}
					break;
				}
				case SCL_ENUMS: {
					String newText = determineTextForPackagePh(cursor.getPlaceholder());
					cursor.getRange().setText(newText);
					newText = cursor.getRange().getText();
					cursor.getPlaceholder().setReplacedText(newText);
					if (isDeep && cursor.getPlaceholder().getErrorText() == null) {
						try {
							cursor = writeSclEnum(cursor, getInput().getPackageDocs().get(newText));
						} catch (Exception e) {
							_logger.warn("#######  Caught exception - skipping: " + e.getMessage());
							e.printStackTrace(System.err);
						}
					}
					break;
				}
				case PACKAGE: {
					String newText = determineTextForPackagePh(cursor.getPlaceholder());
					cursor.getRange().setText(newText);
					newText = cursor.getRange().getText();
					cursor.getPlaceholder().setReplacedText(newText);
					if (isDeep && cursor.getPlaceholder().getErrorText() == null) {
						try {
							PackageDoc packageDoc = getInput().getPackageDocs().get(newText);
							if (packageDoc == null) {
								_logger.warn("No packageDoc for '" + newText + "'.");
								break;
							}
							// this writes all the contents of a root package found as placeholder:
							cursor = writePackage(cursor, packageDoc, true);
						} catch (Exception e) {
							_logger.warn("#######  Caught exception - skipping: " + e.getMessage());
							e.printStackTrace(System.err);
						}
					}
					break;
				}
				case LNMAP_PACKAGE: {
					String newText = determineTextForPackagePh(cursor.getPlaceholder());
					cursor.getRange().setText(newText);
					newText = cursor.getRange().getText();
					cursor.getPlaceholder().setReplacedText(newText);
					if (isDeep && cursor.getPlaceholder().getErrorText() == null) {
						try {
							cursor = writeLnMapPackage(cursor,
									getInput().getPackageDocs().get(newText));
						} catch (Exception e) {
							_logger.warn("#######  Caught exception - skipping: " + e.getMessage());
							e.printStackTrace(System.err);
						}
					}
					break;
				}
				case DATA_INDEX: {
					String newText = determineTextForPackagePh(cursor.getPlaceholder());
					cursor.getRange().setText(newText);
					newText = cursor.getRange().getText();
					cursor.getPlaceholder().setReplacedText(newText);
					if (isDeep && cursor.getPlaceholder().getErrorText() == null) {
						try {
							cursor = writeDataIndex(cursor, getInput().getPackageDocs()
									.get(newText));
						} catch (Exception e) {
							_logger.warn("#######  Caught exception - skipping: " + e.getMessage());
							e.printStackTrace(System.err);
						}
					}
					break;
				}
				case UNSUPPORTED:
				default: {
					String newText = cursor.getPlaceholder().getErrorText();
					cursor.getRange().setText(newText);
					newText = cursor.getRange().getText();
					cursor.getPlaceholder().setReplacedText(newText);
				}
			}

			Util.logCompletion(Level.INFO, String.format("replaced %s...", cursor), start,
					getInput().isSkipTiming());
		}
	}

	private String determineTextForAttributePh(Placeholder ph) {
		if (ph.getErrorText() != null) {
			return ph.getErrorText();
		}

		String attrVal = getInput().getFinder().findAttributeValue(ph.getFirstToken(),
				ph.getSecondToken());
		if (attrVal != null) {
			return attrVal;
		}

		ph.updateModelErrorText();
		return ph.getErrorText();
	}

	private String determineTextForDiagramPhAndInsertDiagram(Range<O> range, Placeholder ph) {
		if (ph.getErrorText() != null) {
			return ph.getErrorText();
		}

		File pic = getInput().getFinder().findDiagramFile(ph.getFirstToken(), ph.getSecondToken());
		if (pic != null) {
			insertFigure(range, pic);
			return "";
		}

		ph.updateModelErrorText();
		return ph.getErrorText();
	}

	// TODO: did we need this? why not used?
	private String determineTextForQualifiedClassPh(Placeholder ph, String separator) {
		if (ph.getErrorText() != null) {
			return ph.getErrorText();
		}

		String qName = getInput().getFinder().findQualifiedClassName(ph.getFirstToken(),
				ph.getSecondToken(), separator);
		if (qName != null) {
			return qName;
		}

		ph.updateModelErrorText();
		return ph.getErrorText();
	}

	private String determineTextForPackagePh(Placeholder ph) {
		if (ph.getErrorText() != null) {
			return ph.getErrorText();
		}

		String pckName = ph.getFirstToken();
		if (getInput().getPackageDocs().keySet().contains(pckName)) {
			return pckName;
		}

		ph.updateModelErrorText();
		return ph.getErrorText();
	}

	// ---------------- actual writing into output document -------------------

	protected static double pointForPerc(int perc) {
		double perc2cm = perc * PAGE_WIDTH / 100.;
		return perc2cm * POINTS_FOR_1CM;
	}

	/**
	 * {@inheritDoc}
	 * <p>
	 * This default implementation does nothing.
	 */
	@Override
	public void clearUndoCache() {
		// no-op
	}

	// ===== Impl. of org.tanjakostic.jcleancim.docgen.writer.word.WordHelper methods =====

	private CursorList<O> _cursors;
	private int _closeReopenEveryCounter = 0;

	@Override
	public final void scanAndPostprocess(PostProcessor pp) {
		doIt(true, pp);
	}

	@Override
	public final CursorList<O> getCursors() {
		return _cursors;
	}

	@Override
	public final boolean isInTOC(Range<O> range) {
		return Style.isTOC(getRangeParagraphStyleName(range, 1));
	}

	@Override
	public final Cursor<O> createCursor(Placeholder ph, Range<O> limited) {
		return new Cursor<O>(ph, limited);
	}

	@Override
	public final Caption<O> createCaption(CaptionKind figure, Range<O> range) {
		return new Caption<O>(figure, range);
	}

	@Override
	public CursorList<O> initCursorsFromPlaceholders(List<Range<O>> figCaptionRanges,
			List<Range<O>> tabCaptionRanges) {
		CursorList<O> result = new CursorList<O>();

		int figCountBefore = 0;
		int tabCountBefore = 0;
		int i = 0;

		// String msPattern = Placeholder.MS_PATTERN;
		WordPatternFinder<O> finder = createPatternFinder();
		Range<O> range = finder.getRange();

		// Range<Dispatch> range = createRange(docContent);
		// Dispatch find = initFinder(docContent, msPattern);

		while (finder.hasMore()) {
			String foundPattern = range.getText();

			if (isInTOC(range)) {
				_logger.info("   skipping TOC: " + range.getText());
				continue;
			}

			if (i != 0) {
				Cursor<O> prevCursor = result.get(i - 1);
				Range<O> prevRange = prevCursor.getRange();
				figCountBefore = prevCursor.getPlaceholder().getFigureCount();
				for (Range<O> capRange : figCaptionRanges) {
					if (prevRange.getEnd() < capRange.getStart()
							&& capRange.getEnd() < range.getStart()) {
						figCountBefore++;
					}
				}

				tabCountBefore = prevCursor.getPlaceholder().getTableCount();
				for (Range<O> capRange : tabCaptionRanges) {
					if (prevRange.getEnd() < capRange.getStart()
							&& capRange.getEnd() < range.getStart()) {
						tabCountBefore++;
					}
				}
			} else {
				figCountBefore = 0;
				for (Range<O> capRange : figCaptionRanges) {
					if (capRange.getEnd() < range.getStart()) {
						figCountBefore++;
					}
				}

				tabCountBefore = 0;
				for (Range<O> capRange : tabCaptionRanges) {
					if (capRange.getEnd() < range.getStart()) {
						tabCountBefore++;
					}
				}
			}

			Placeholder ph = new Placeholder(foundPattern, figCountBefore, tabCountBefore);
			Cursor<O> cursor = createCursor(ph, duplicateRange(range));
			result.add(cursor);
			_logger.info("   saved placeholder '" + range.getText() + "' " + ph.toString());
			++i;
		}
		return result;
	}

	/**
	 * {@inheritDoc}
	 * <p>
	 * This default implementation just returns <code>currentCursor</code>, without any
	 * closing/reopening. If you need to actually close/reopen the document (as a means of
	 * optimising performance), override this method.
	 */
	@Override
	public Cursor<O> closeAndReopenDoc(CursorList<O> cursors, Cursor<O> currentCursor) {
		return currentCursor;
	}

	/**
	 * You'll always use this one for regular text and tables.
	 */
	@Override
	public String appendTextInNewParagraphWithStyle(Range<O> range, TextDescription newText,
			Style style) {
		boolean isHtml = newText.kind == TextKind.htmlSnippet;
		return (isHtml) ? appendHtmlTextInNewParagraphWithStyle(range, newText.text, style)
				: appendRawTextInNewParagraphWithStyle(range, newText.text, style);
	}

	// ===== Impl. of org.tanjakostic.jcleancim.docgen.writer.word.WordWriter methods =====

	@Override
	public Cursor<O> writePackage(Cursor<O> initCursor, PackageDoc doc, boolean isRoot) {
		Cursor<O> cursor = initCursor;

		Util.ensureNotNull(cursor, "cursor");
		Util.ensureNotNull(doc, "doc");
		if (isRoot) {
			String headingText = doc.getHeadingText();
			if (!headingText.equals(cursor.getRange().getText())) {
				cursor.getRange().setText(headingText);
			}
		}

		int outlineLevel = getRangeParagraphOutlineLevel(cursor.getRange(), 1);
		_logger.info("   writing doc for package " + Util.getIndentSpaces(outlineLevel)
				+ doc.getPackageName() + " ...");

		Style genHeadStyle = Style.getHeadingStyle(outlineLevel + 1);
		if (doc.getClassDocs().size() > 0 || doc.getChildPackageDocs().size() > 0) {
			appendTextInNewParagraphWithStyle(cursor.getRange(),
					new TextDescription(doc.getGenHeadingText()), genHeadStyle);
		}
		appendTextInNewParagraphWithStyle(cursor.getRange(), doc.getDescription(), Style.para);

		for (int i = 0; i < doc.getFigureDocs().size(); ++i) {
			cursor = writeDiagram(cursor, doc.getFigureDocs().get(i));
		}

		for (int i = 0; i < doc.getClassDocs().size(); ++i) {
			ClassDoc cDoc = doc.getClassDocs().get(i);
			appendTextInNewParagraphWithStyle(cursor.getRange(),
					new TextDescription(cDoc.getHeadingText()), genHeadStyle);
			cursor = writeClass(cursor, cDoc);
		}
		for (int i = 0; i < doc.getChildPackageDocs().size(); ++i) {
			PackageDoc pDoc = doc.getChildPackageDocs().get(i);
			appendTextInNewParagraphWithStyle(cursor.getRange(),
					new TextDescription(pDoc.getHeadingText()), genHeadStyle);
			cursor = writePackage(cursor, pDoc, false);
		}
		return cursor;
	}

	@Override
	public Cursor<O> writeDataIndex(Cursor<O> initCursor, PackageDoc packageDoc) {
		Cursor<O> cursor = initCursor;
		String packageName = packageDoc.getPackageName();
		PropertiesDoc doc = packageDoc.getDataIndexDoc();
		if (doc == null) {
			_logger.error("Data index doc null for package " + packageName
					+ " - template contains placeholder " + cursor.getPlaceholder().getText() + ","
					+ " whereas property '" + Config.KEY_VALIDATION_PACKAGES_DATA_INDEX
					+ "' does not contain package '" + packageName + "'.");
			return cursor;
		}

		_logger.info("   writing data index doc for package " + packageName + " ...");
		cursor.getRange().setText(doc.getHeadingText());
		cursor = writeProperties(cursor, doc);
		return cursor;
	}

	@Override
	public Cursor<O> writeLnMapPackage(Cursor<O> initCursor, PackageDoc packageDoc) {
		Cursor<O> cursor = initCursor;
		String packageName = packageDoc.getPackageName();
		PropertiesDoc doc = packageDoc.getLnMapPackageDoc();
		if (doc == null) {
			_logger.error("LN mappings doc null for package " + packageName + " - property '"
					+ Config.KEY_VALIDATION_IEC61850_PACKAGE_LN_MAPS + "' not specified or empty.");
			return cursor;
		}

		_logger.info("   writing LN mappings doc for package " + packageName + " ...");
		cursor.getRange().setText(doc.getHeadingText());
		appendTextInNewParagraphWithStyle(cursor.getRange(), doc.getDescription(), Style.para);
		cursor = writeProperties(cursor, doc);
		return cursor;
	}

	@Override
	public Cursor<O> writePresCondTable(Cursor<O> initCursor, PackageDoc packageDoc) {
		Cursor<O> cursor = initCursor;
		String packageName = packageDoc.getPackageName();
		PropertiesDoc doc = packageDoc.getPresCondPackageDoc();
		if (doc == null) {
			_logger.error("Presence conditions doc null for package " + packageName
					+ " - property '" + Config.KEY_VALIDATION_IEC61850_PACKAGE_PRES_COND
					+ "' not specified or empty.");
			return cursor;
		}

		_logger.info("   writing presence conditions from package " + packageName + " ...");
		cursor.getRange().setText(doc.getHeadingText());
		appendTextInNewParagraphWithStyle(cursor.getRange(), doc.getDescription(), Style.para);
		cursor = writeProperties(cursor, doc);
		return cursor;
	}

	@Override
	public Cursor<O> writeFcTable(Cursor<O> initCursor, PackageDoc packageDoc) {
		Cursor<O> cursor = initCursor;
		String packageName = packageDoc.getPackageName();
		PropertiesDoc doc = packageDoc.getFcPackageDoc();
		if (doc == null) {
			_logger.error("Functional constraints doc null for package " + packageName
					+ " - property '" + Config.KEY_VALIDATION_IEC61850_PACKAGE_FC
					+ "' not specified or empty.");
			return cursor;
		}

		_logger.info("   writing functional constraints from package " + packageName + " ...");
		cursor.getRange().setText(doc.getHeadingText());
		appendTextInNewParagraphWithStyle(cursor.getRange(), doc.getDescription(), Style.para);
		cursor = writeProperties(cursor, doc);
		return cursor;
	}

	@Override
	public Cursor<O> writeTrgOpTable(Cursor<O> initCursor, PackageDoc packageDoc) {
		Cursor<O> cursor = initCursor;
		String packageName = packageDoc.getPackageName();
		PropertiesDoc doc = packageDoc.getTrgOpPackageDoc();
		if (doc == null) {
			_logger.error("Trigger options doc null for package " + packageName + " - property '"
					+ Config.KEY_VALIDATION_IEC61850_PACKAGE_TRGOP + "' not specified or empty.");
			return cursor;
		}

		_logger.info("   writing trigger options from package " + packageName + " ...");
		cursor.getRange().setText(doc.getHeadingText());
		appendTextInNewParagraphWithStyle(cursor.getRange(), doc.getDescription(), Style.para);
		cursor = writeProperties(cursor, doc);
		return cursor;
	}

	@Override
	public Cursor<O> writeAbbrTable(Cursor<O> initCursor, PackageDoc packageDoc) {
		Cursor<O> cursor = initCursor;
		String packageName = packageDoc.getPackageName();
		PropertiesDoc doc = packageDoc.getAbbrPackageDoc();
		if (doc == null) {
			_logger.error("Abbreviations doc null for package " + packageName + " - property '"
					+ Config.KEY_VALIDATION_IEC61850_PACKAGES_DO_ABBR + "' not specified or empty.");
			return cursor;
		}

		_logger.info("   writing abbreviations from package " + packageName + " ...");
		cursor.getRange().setText("");
		cursor = writeProperties(cursor, doc);
		return cursor;
	}

	@Override
	public Cursor<O> writeSclEnum(Cursor<O> cursor, PackageDoc packageDoc) {
		String packageName = packageDoc.getPackageName();
		PackageScl doc = packageDoc.getEnumsPackageScl();
		if (doc == null) {
			_logger.error("Enum XML doc null for package " + packageName + " - property '"
					+ Config.KEY_VALIDATION_IEC61850_PACKAGES_ENUMS_XML
					+ "' not specified or empty.");
			return cursor;
		}

		_logger.info("   writing SCL for enums from package " + packageName + " ...");
		cursor.getRange().setText(doc.getHeadingText());
		appendRawTextInNewParagraphWithStyle(cursor.getRange(), doc.toString(), Style.tabcell);
		return cursor;
	}

	@Override
	public Cursor<O> writeClass(Cursor<O> initCursor, ClassDoc doc) {
		Cursor<O> cursor = initCursor;

		clearUndoCache(); // to avoid word showing "memory insufficient..."

		appendTextInNewParagraphWithStyle(cursor.getRange(), doc.getDescription(), Style.para);

		for (int i = 0; i < doc.getDiagramDocs().size(); ++i) {
			cursor = writeDiagram(cursor, doc.getDiagramDocs().get(i));
		}

		if (doc.getAttributesDoc().notEmpty()) {
			cursor = writeProperties(cursor, doc.getAttributesDoc());
		}
		if (doc.getAssocEndsDoc().notEmpty()) {
			cursor = writeProperties(cursor, doc.getAssocEndsDoc());
		}
		if (doc.getOperationsDoc().notEmpty()) {
			cursor = writeProperties(cursor, doc.getOperationsDoc());
		}

		return cursor;
	}

	@Override
	public Cursor<O> writeProperties(Cursor<O> initCursor, PropertiesDoc doc) {
		Cursor<O> cursor = initCursor;
		if (_closeReopenEveryCounter == getInput().getSaveRecloseEvery()) {
			_closeReopenEveryCounter = 0;
			cursor = closeAndReopenDoc(getCursors(), cursor);
		}

		Range<O> range = cursor.getRange();
		int insertStart = range.getEnd() + 1; // selects EOL as well

		appendTextInNewParagraphWithStyle(range, new TextDescription(doc.getIntroText()),
				Style.para);

		Range<O> insertPoint = duplicateRange(range);
		insertPoint.setStart(insertStart);

		insertTable(range, doc, Style.tabhead);

		int tabNumber = getCursors().captionAdded(CaptionKind.Table, cursor);
		insertTableCaption(range, doc.getCaptionText(), ("Table " + tabNumber));

		insertTableRef(insertPoint, tabNumber);

		++_closeReopenEveryCounter;
		return cursor;
	}

	@Override
	public Cursor<O> writeDiagram(Cursor<O> cursor, FigureDoc doc) {
		Range<O> range = cursor.getRange();
		int insertStart = range.getEnd() + 1; // selects EOL as well

		if (_input.isIntroToFigureBefore()) {
			appendTextInNewParagraphWithStyle(range, new TextDescription(doc.getIntroText()),
					Style.para);
		}

		Range<O> insertPoint = duplicateRange(range);
		insertPoint.setStart(insertStart);

		appendTextInNewParagraphWithStyle(range, new TextDescription(), Style.fig);
		insertFigure(range, doc.getFigureFile());

		int figNumber = getCursors().captionAdded(CaptionKind.Figure, cursor);
		String figRefTxt = "Figure " + figNumber;
		insertFigureCaption(range, doc.getCaptionText(), figRefTxt);

		TextDescription txtBelowFigure = doc.getDescription();
		if (_input.isIntroToFigureBefore()) {
			insertFigureRef(insertPoint, figNumber);
		} else {
			txtBelowFigure = txtBelowFigure.prepend(figRefTxt + ": ");
		}
		appendTextInNewParagraphWithStyle(range, txtBelowFigure, Style.para);

		range.setStart(insertStart);

		return cursor;
	}

	// ===== Impl. of org.tanjakostic.jcleancim.docgen.writer.Writer methods =====

	@Override
	public final WordWriterInput getInput() {
		return _input;
	}

	@Override
	public final String getInputFileNames() {
		return getInput().getInTemplatePath();
	}

	@Override
	public final String getOutputFileNames() {
		return getInput().getOutFilePath();
	}

	@Override
	public final void write() {
		doIt(false, null);
	}

	@Override
	public String toString() {
		return String.format("%s [modelFileName=%s, model=%s, outFilePath=%s]", getClass()
				.getSimpleName(), getInput().getModelFileName(), getInput().getFinder(), getInput()
				.getOutFilePath());
	}
}
