/**
 * Copyright (C) 2009-2015 Tatjana (Tanja) Kostic
 * <p>
 * This file belongs to jCleanCim, a tool supporting tasks of UML model managers for IEC TC57 CIM
 * and 61850 models.
 * <p>
 * This program is free software: you can redistribute it and/or modify it under the terms of the
 * GNU Lesser General Public License version 3.
 * <p>
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details. You should have received a copy of the GNU Lesser
 * General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

package org.tanjakostic.jcleancim.docgen.writer.word;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;

import org.tanjakostic.jcleancim.common.Config;
import org.tanjakostic.jcleancim.docgen.collector.FreeFormDocumentation;
import org.tanjakostic.jcleancim.docgen.collector.ModelFinder;
import org.tanjakostic.jcleancim.docgen.collector.PackageDoc;
import org.tanjakostic.jcleancim.docgen.writer.WriterInput;
import org.tanjakostic.jcleancim.util.Util;

/**
 * @author tatjana.kostic@ieee.org
 * @version $Id: WordWriterInput.java 2728 2015-05-14 20:00:03Z CHTAKOS $
 */
public class WordWriterInput extends WriterInput {

	private final Map<String, PackageDoc> _packageDocs;
	private final ModelFinder _finder;
	private final String _inTemplatePath;
	private final String _outFilePath;
	private final boolean _useBinaryDoc;
	private final boolean _introToFigureBefore;
	private final int _saveRecloseEvery;
	private final boolean _deep;

	/**
	 * Constructor.
	 *
	 * @param cfg
	 * @param freeFormDoc
	 *            "flattened" map of package documentation instances, with package name as key (to
	 *            allow to quickly find the package name from what is read in the placeholder). If
	 *            null or empty, placeholders dealing with packages will all have error and empty
	 *            content.
	 */
	public WordWriterInput(Config cfg, FreeFormDocumentation freeFormDoc) {
		this(cfg, null, cfg.isAppSkipTiming(), null, freeFormDoc.getPackageDocs(), freeFormDoc
				.getModelFinder(), cfg.getDocgenWordInTemplateFileAbsPath(), cfg
				.getDocgenWordOutDocumentFileAbsPath(), cfg.isDocgenWordUseDocFormat(), cfg
				.isDocgenWordIntroToFigureBefore(), cfg.getDocgenWordSaveReopenEvery(), !cfg
				.isDocgenWordAnalysePlaceholders());
	}

	/**
	 * Constructor, useful to create an instance if you don't have configuration.
	 *
	 * @param appVersion
	 * @param modelFileName
	 * @param skipTiming
	 * @param packageDocs
	 *            "flattened" map of package documentation instances, with package name as key (to
	 *            allow to quickly find the package name from what is read in the placeholder). If
	 *            null or empty, placeholders dealing with packages will all have error and empty
	 *            content.
	 * @param finder
	 *            model facade; if null, most of placeholders will have error.
	 * @param inTemplatePath
	 * @param outFilePath
	 * @param useBinaryDoc
	 * @param introToFigureBefore
	 * @param saveRecloseEvery
	 * @param isDeep
	 */
	public WordWriterInput(String appVersion, String modelFileName, boolean skipTiming,
			Map<String, PackageDoc> packageDocs, ModelFinder finder, String inTemplatePath,
			String outFilePath, boolean useBinaryDoc, boolean introToFigureBefore,
			int saveRecloseEvery, boolean isDeep) {
		this(null, appVersion, skipTiming, modelFileName, packageDocs, finder, inTemplatePath,
				outFilePath, useBinaryDoc, introToFigureBefore, saveRecloseEvery, isDeep);
	}

	private WordWriterInput(Config cfg, String appVersion, boolean skipTiming,
			String modelFileName, Map<String, PackageDoc> packageDocs, ModelFinder finder,
			String inTemplatePath, String outFilePath, boolean useBinaryDoc,
			boolean introToFigureBefore, int saveRecloseEvery, boolean isDeep) {
		super(cfg, appVersion, modelFileName, skipTiming);

		Util.ensureNotEmpty(outFilePath, "outFilePath");

		_packageDocs = (packageDocs == null || packageDocs.isEmpty()) ? new LinkedHashMap<String, PackageDoc>()
				: packageDocs;

		_finder = finder;
		_inTemplatePath = Util.null2empty(inTemplatePath);
		_outFilePath = outFilePath;
		_useBinaryDoc = useBinaryDoc;
		_introToFigureBefore = introToFigureBefore;
		_saveRecloseEvery = saveRecloseEvery;
		_deep = isDeep;
	}

	/**
	 * Returns "flattened" map of package documentation instances, with package name as key (to
	 * allow to quickly find the package name from what is read in the placeholder). If null or
	 * empty, placeholders dealing with packages will by definition all report error.
	 */
	public final Map<String, PackageDoc> getPackageDocs() {
		return Collections.unmodifiableMap(_packageDocs);
	}

	/**
	 * Returns object that can find in the UML model items specified in placeholders. If null, most
	 * of placeholders will by definition report error.
	 */
	public ModelFinder getFinder() {
		return _finder;
	}

	/**
	 * Returns absolute path of the file used as template for documentation, empty string if
	 * template is not used.
	 */
	public String getInTemplatePath() {
		return _inTemplatePath;
	}

	/** Returns absolute path of the file to which to write documentation. */
	public String getOutFilePath() {
		return _outFilePath;
	}

	/** Returns whether to force use of COM API (binary .doc format). */
	public boolean isUseBinaryDoc() {
		return _useBinaryDoc;
	}

	/** Returns whether to force figure introduction sentence before the figure. */
	public boolean isIntroToFigureBefore() {
		return _introToFigureBefore;
	}

	/**
	 * Returns the number of tables (captions) to print before closing and reopening the file; this
	 * is an optimisation option that may not be applicable to all writers.
	 */
	public int getSaveRecloseEvery() {
		return _saveRecloseEvery;
	}

	/**
	 * Returns whether to write content for UML packages. Value false is useful for analysing
	 * placeholders in the template document, without writing the whole content.
	 */
	public boolean isDeep() {
		return _deep;
	}
}
