/**
 * Copyright (C) 2009-2015 Tatjana (Tanja) Kostic
 * <p>
 * This file belongs to jCleanCim, a tool supporting tasks of UML model managers for IEC TC57 CIM
 * and 61850 models.
 * <p>
 * This program is free software: you can redistribute it and/or modify it under the terms of the
 * GNU Lesser General Public License version 3.
 * <p>
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details. You should have received a copy of the GNU Lesser
 * General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

package org.tanjakostic.jcleancim.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;
import org.tanjakostic.jcleancim.util.Util;

/**
 * Content of namespace class. In case of CIM, the information is deduced from version class, while
 * for IEC61850, the information is extracted from the namespace class.
 *
 * @author tatjana.kostic@ieee.org
 * @version $Id: NamespaceInfo.java 2728 2015-05-14 20:00:03Z CHTAKOS $
 */
public class NamespaceInfo {
	private static final Logger _logger = Logger.getLogger(NamespaceInfo.class.getName());

	private final String _id;
	private final String _version;
	private final String _revision;
	private final String _date;
	private final String _umlVersion;
	private final String _tissuesApplied;
	private final Set<NamespaceInfo> _dependencies = new LinkedHashSet<NamespaceInfo>();

	/**
	 * Constructs instance with empty revision and tissues, and with UML version same as namespace
	 * version; this is for CIM namespaces, which are deduced from UML version class.
	 *
	 * @param id
	 * @param version
	 * @param date
	 */
	public NamespaceInfo(String id, String version, String date) {
		this(id, version, "", date, version, "");
	}

	/**
	 * Constructor for IEC61850 namespaces.
	 *
	 * @param id
	 * @param version
	 * @param revision
	 * @param date
	 * @param umlVersion
	 * @param tissuesApplied
	 */
	public NamespaceInfo(String id, String version, String revision, String date,
			String umlVersion, String tissuesApplied) {
		_id = id;
		_version = version;
		_revision = revision;
		_date = date;
		_umlVersion = umlVersion;
		_tissuesApplied = tissuesApplied;
	}

	public String getId() {
		return _id;
	}

	public String getVersion() {
		return _version;
	}

	public String getRevision() {
		return _revision;
	}

	public String getDate() {
		return _date;
	}

	public String getUmlVersion() {
		return _umlVersion;
	}

	public String getTissuesApplied() {
		return _tissuesApplied;
	}

	/** Returns all dependencies of this namespace. */
	public Set<NamespaceInfo> getDependencies() {
		return Collections.unmodifiableSet(_dependencies);
	}

	/**
	 * Adds <code>namespace</code> as dependency to this namespace, if the dependency is not cirular
	 * and returns whether addition happened.
	 */
	public boolean addDependency(NamespaceInfo namespace) {
		Util.ensureNotNull(namespace, "namespace");
		if (namespace == this) {
			_logger.warn(String.format("Cannot add self (%s) as dependency !",
					namespace.getNamespace()));
			return false;
		}
		if (namespace.getDependencies().contains(this)) {
			_logger.warn(String.format("Cannot add namespace (%s) with circular"
					+ " dependency to me (%s) !", namespace.getNamespace(), getNamespace()));
			return false;
		}
		return _dependencies.add(namespace);
	}

	/** Returns formatted string including id, version and revision. */
	public String getNamespace() {
		return String.format("%s:%s%s", getId(), getVersion(), getRevision());
	}

	/** Returns namespace strings of all the dependencies. */
	public List<String> getDependencyStrings() {
		List<String> result = new ArrayList<String>();
		for (NamespaceInfo ns : getDependencies()) {
			result.add(ns.getNamespace());
		}
		return result;
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("namespace ").append(getNamespace()).append(", ").append(getDate());
		if (!getUmlVersion().isEmpty()) {
			sb.append(";  UML version ").append(getUmlVersion());
		}
		if (!getTissuesApplied().isEmpty()) {
			sb.append(";  tissues applied: ").append(getTissuesApplied());
		}
		if (!getDependencies().isEmpty()) {
			sb.append(";  requires: ").append(
					Util.concatCharSeparatedTokens(", ", true, getDependencyStrings()));
		}
		return sb.toString();
	}
}
