/**
 * Redistribution and use in source and binary forms, with or without modification, are permitted.
 * <p>
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR ONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.tanjakostic.jcleancim.docgen.writer;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertSame;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.tanjakostic.jcleancim.docgen.writer.Placeholder;
import org.tanjakostic.jcleancim.docgen.writer.Placeholder.Kind;

/**
 * @author tatjana.kostic@ieee.org
 * @version $Id: PlaceholderTest.java 2728 2015-05-14 20:00:03Z CHTAKOS $
 */
public class PlaceholderTest {

	static final String CLASS_NAME = "className";
	static final String DIAGRAM_NAME = "diagramName";
	static final String ATTR_NAME = "attrName";
	static final String PCKG_NAME = "pckgName";
	static final String FILE_NAME = "fileName";

	@Before
	public void setUp() {
		// nothing
	}

	@After
	public void tearDown() {
		// nothing
	}

	static Placeholder createUnsupportedInstance() {
		return new Placeholder("invalidPlaceholder");
	}

	static Placeholder createFileInstance() {
		String text = Placeholder.constructFilePlaceholderText();
		return new Placeholder(text);
	}

	static Placeholder createDiagramInstance() {
		String text = Placeholder.constructDiagramPlaceholderText(PCKG_NAME, DIAGRAM_NAME);
		return new Placeholder(text);
	}

	static Placeholder createAttributeInstance() {
		String text = Placeholder.constructAttributePlaceholderText(CLASS_NAME, ATTR_NAME);
		return new Placeholder(text);
	}

	static Placeholder createPresConditionsPackageInstance() {
		String text = Placeholder.constructPresConditionsPackagePlaceholderText(PCKG_NAME);
		return new Placeholder(text);
	}

	static Placeholder createFcPackageInstance() {
		String text = Placeholder.constructFcsPackagePlaceholderText(PCKG_NAME);
		return new Placeholder(text);
	}

	static Placeholder createTrgOpPackageInstance() {
		String text = Placeholder.constructTrgOpsPackagePlaceholderText(PCKG_NAME);
		return new Placeholder(text);
	}

	static Placeholder createAbbrPackageInstance() {
		String text = Placeholder.constructAbbrPackagePlaceholderText(PCKG_NAME);
		return new Placeholder(text);
	}

	static Placeholder createEnumPackageInstance() {
		String text = Placeholder.constructEnumPackagePlaceholderText(PCKG_NAME);
		return new Placeholder(text);
	}

	static Placeholder createPackageInstance() {
		String text = Placeholder.constructPackagePlaceholderText(PCKG_NAME);
		return new Placeholder(text);
	}

	static Placeholder createPackageInstance(int figCount, int tabCount) {
		String text = Placeholder.constructPackagePlaceholderText(PCKG_NAME);
		return new Placeholder(text, figCount, tabCount);
	}

	static Placeholder createSpecialPackageInstance() {
		String text = Placeholder.constructLNMapPackagePlaceholderText(PCKG_NAME);
		return new Placeholder(text);
	}

	static Placeholder createDataIndexInstance() {
		String text = Placeholder.constructDataIndexPlaceholderText(PCKG_NAME);
		return new Placeholder(text);
	}

	// ============= Tests ===============

	@Test
	public void testCtor_unsupported() {
		Placeholder ph = createUnsupportedInstance();
		assertNull(ph.getFirstToken());
		assertNull(ph.getSecondToken());
		assertNotNull(ph.getErrorText());
		assertSame(Kind.UNSUPPORTED, ph.getKind());
	}

	@Test
	public void testCtor_file() {
		Placeholder ph = createFileInstance();
		assertNull(ph.getFirstToken());
		assertNull(ph.getSecondToken());
		assertNull(ph.getErrorText());
		assertSame(Kind.FILE, ph.getKind());
	}

	@Test
	public void testCtor_attr() {
		Placeholder ph = createAttributeInstance();
		assertEquals(CLASS_NAME, ph.getFirstToken());
		assertEquals(ATTR_NAME, ph.getSecondToken());
		assertNull(ph.getErrorText());
		assertSame(Kind.ATTRIBUTE, ph.getKind());
	}

	@Test
	public void testCtor_diagram() {
		Placeholder ph = createDiagramInstance();
		assertEquals(PCKG_NAME, ph.getFirstToken());
		assertEquals(DIAGRAM_NAME, ph.getSecondToken());
		assertNull(ph.getErrorText());
		assertSame(Kind.DIAGRAM, ph.getKind());
	}

	@Test
	public void testCtor_presConditions() {
		Placeholder ph = createPresConditionsPackageInstance();
		assertEquals(PCKG_NAME, ph.getFirstToken());
		assertNull(ph.getSecondToken());
		assertNull(ph.getErrorText());
		assertSame(Kind.PRES_CONDITIONS, ph.getKind());
	}

	@Test
	public void testCtor_fcs() {
		Placeholder ph = createFcPackageInstance();
		assertEquals(PCKG_NAME, ph.getFirstToken());
		assertNull(ph.getSecondToken());
		assertNull(ph.getErrorText());
		assertSame(Kind.FCS, ph.getKind());
	}

	@Test
	public void testCtor_trgops() {
		Placeholder ph = createTrgOpPackageInstance();
		assertEquals(PCKG_NAME, ph.getFirstToken());
		assertNull(ph.getSecondToken());
		assertNull(ph.getErrorText());
		assertSame(Kind.TRGOPS, ph.getKind());
	}

	@Test
	public void testCtor_abbreviations() {
		Placeholder ph = createAbbrPackageInstance();
		assertEquals(PCKG_NAME, ph.getFirstToken());
		assertNull(ph.getSecondToken());
		assertNull(ph.getErrorText());
		assertSame(Kind.ABBREVIATIONS, ph.getKind());
	}

	@Test
	public void testCtor_enumerations() {
		Placeholder ph = createEnumPackageInstance();
		assertEquals(PCKG_NAME, ph.getFirstToken());
		assertNull(ph.getSecondToken());
		assertNull(ph.getErrorText());
		assertSame(Kind.SCL_ENUMS, ph.getKind());
	}

	@Test
	public void testCtor_package() {
		Placeholder ph = createPackageInstance();
		assertEquals(PCKG_NAME, ph.getFirstToken());
		assertNull(ph.getSecondToken());
		assertNull(ph.getErrorText());
		assertSame(Kind.PACKAGE, ph.getKind());
	}

	@Test
	public void testCtor_packageWithFigsTabs() {
		int figCount = 2;
		int tabCount = 1;
		Placeholder ph = createPackageInstance(figCount, tabCount);
		assertEquals(figCount, ph.getFigureCountBefore());
		assertEquals(tabCount, ph.getTableCountBefore());
	}

	@Test
	public void testCtor_specialPackage() {
		Placeholder ph = createSpecialPackageInstance();
		assertEquals(PCKG_NAME, ph.getFirstToken());
		assertNull(ph.getSecondToken());
		assertNull(ph.getErrorText());
		assertSame(Kind.LNMAP_PACKAGE, ph.getKind());
	}

	@Test
	public void testCtor_dataIndex() {
		Placeholder ph = createDataIndexInstance();
		assertEquals(PCKG_NAME, ph.getFirstToken());
		assertNull(ph.getSecondToken());
		assertNull(ph.getErrorText());
		assertSame(Kind.DATA_INDEX, ph.getKind());
	}

	// -----------------------------

	@Test
	public void testGetErrorText_abbrPckTokenEmpty() {
		String text = Placeholder.constructAbbrPackagePlaceholderText("");
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	@Test
	public void testGetErrorText_abbrPckTokenEmptyNull() {
		String text = Placeholder.constructAbbrPackagePlaceholderText(null);
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	// -----------------------------

	@Test
	public void testGetErrorText_presCondPckTokenEmpty() {
		String text = Placeholder.constructPresConditionsPackagePlaceholderText("");
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	@Test
	public void testGetErrorText_presCondPckTokenEmptyNull() {
		String text = Placeholder.constructPresConditionsPackagePlaceholderText(null);
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	// -----------------------------

	@Test
	public void testGetErrorText_presFcPckTokenEmpty() {
		String text = Placeholder.constructFcsPackagePlaceholderText("");
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	@Test
	public void testGetErrorText_presFcPckTokenEmptyNull() {
		String text = Placeholder.constructFcsPackagePlaceholderText(null);
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	// -----------------------------

	@Test
	public void testGetErrorText_presTrgOpPckTokenEmpty() {
		String text = Placeholder.constructTrgOpsPackagePlaceholderText("");
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	@Test
	public void testGetErrorText_presTrgOpPckTokenEmptyNull() {
		String text = Placeholder.constructTrgOpsPackagePlaceholderText(null);
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	// -----------------------------

	@Test
	public void testGetErrorText_enumPckTokenEmpty() {
		String text = Placeholder.constructEnumPackagePlaceholderText("");
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	@Test
	public void testGetErrorText_enumPckTokenEmptyNull() {
		String text = Placeholder.constructEnumPackagePlaceholderText(null);
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	// -----------------------------

	@Test
	public void testGetErrorText_pckTokenEmpty() {
		String text = Placeholder.constructPackagePlaceholderText("");
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	@Test
	public void testGetErrorText_pckTokenEmptyNull() {
		String text = Placeholder.constructPackagePlaceholderText(null);
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	// -----------------------------

	@Test
	public void testGetErrorText_spPckTokenEmpty() {
		String text = Placeholder.constructLNMapPackagePlaceholderText("");
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	@Test
	public void testGetErrorText_spPckTokenEmptyNull() {
		String text = Placeholder.constructLNMapPackagePlaceholderText(null);
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	// -----------------------------

	@Test
	public void testGetErrorText_diTokenEmpty() {
		String text = Placeholder.constructDataIndexPlaceholderText("");
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	@Test
	public void testGetErrorText_diTokenEmptyNull() {
		String text = Placeholder.constructDataIndexPlaceholderText(null);
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	// -----------------

	@Test
	public void testGetErrorText_attrFirstTokenEmpty() {
		String text = Placeholder.constructAttributePlaceholderText("", ATTR_NAME);
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	@Test
	public void testGetErrorText_attrFirstTokenNull() {
		String text = Placeholder.constructAttributePlaceholderText(null, ATTR_NAME);
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	@Test
	public void testGetErrorText_attrSecTokenEmpty() {
		String text = Placeholder.constructAttributePlaceholderText(CLASS_NAME, "");
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	@Test
	public void testGetErrorText_attrSecTokenNull() {
		String text = Placeholder.constructAttributePlaceholderText(CLASS_NAME, "");
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	// -----------------

	@Test
	public void testGetErrorText_diagFirstTokenEmpty() {
		String text = Placeholder.constructAttributePlaceholderText("", DIAGRAM_NAME);
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	@Test
	public void testGetErrorText_diagFirstTokenNull() {
		String text = Placeholder.constructAttributePlaceholderText(null, DIAGRAM_NAME);
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	@Test
	public void testGetErrorText_diagSecTokenEmpty() {
		String text = Placeholder.constructAttributePlaceholderText(PCKG_NAME, "");
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	@Test
	public void testGetErrorText_diagSecTokenNull() {
		String text = Placeholder.constructAttributePlaceholderText(PCKG_NAME, "");
		Placeholder ph = new Placeholder(text);
		assertNotNull(ph.getErrorText());
	}

	// -----------------

	@Test
	public void testAddFigureAddTable() {
		int figCount = 2;
		int tabCount = 1;
		Placeholder ph = createPackageInstance(figCount, tabCount);
		assertEquals(figCount, ph.getFigureCountBefore());
		assertEquals(tabCount, ph.getTableCountBefore());

		assertEquals(figCount + 1, ph.addFigure());
		assertEquals(tabCount + 1, ph.addTable());
	}

	@Test
	public void testGetFigureAndTableCount() {
		int figCount = 2;
		int tabCount = 1;
		Placeholder ph = createPackageInstance(figCount, tabCount);
		assertEquals(figCount, ph.getFigureCount());
		assertEquals(tabCount, ph.getTableCount());

		ph.addFigure();
		ph.addTable();

		assertEquals(figCount + 1, ph.getFigureCount());
		assertEquals(tabCount + 1, ph.getTableCount());
	}

	@Test
	public void testIncrementFigureAndTableBefore() {
		Placeholder ph = createPackageInstance(2, 1);
		assertEquals(2, ph.getFigureCount());
		assertEquals(1, ph.getTableCount());

		ph.addFigure();
		ph.addTable();

		assertEquals(3, ph.getFigureCount());
		assertEquals(2, ph.getTableCount());

		ph.incrementFigureBefore();
		ph.incrementTableBefore();

		assertEquals(3, ph.getFigureCountBefore());
		assertEquals(2, ph.getTableCountBefore());
		assertEquals(4, ph.getFigureCount());
		assertEquals(3, ph.getTableCount());
	}
}
