/**
 * Redistribution and use in source and binary forms, with or without modification, are permitted.
 * <p>
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR ONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.tanjakostic.jcleancim.model;

import static org.junit.Assert.assertSame;
import static org.junit.Assert.assertTrue;

import java.util.Properties;

import org.junit.After;
import org.junit.Before;
import org.tanjakostic.jcleancim.common.Config;
import org.tanjakostic.jcleancim.common.Nature;
import org.tanjakostic.jcleancim.model.UmlAssociation;
import org.tanjakostic.jcleancim.model.UmlAssociationEnd;
import org.tanjakostic.jcleancim.model.UmlClass;
import org.tanjakostic.jcleancim.model.UmlModel;
import org.tanjakostic.jcleancim.model.UmlObjectData;
import org.tanjakostic.jcleancim.model.UmlPackage;
import org.tanjakostic.jcleancim.model.UmlStereotype;
import org.tanjakostic.jcleancim.util.ApplicationException;

/**
 * Common fixture and utility methods.
 * 
 * @author tatjana.kostic@ieee.org
 * @version $Id: CommonUmlTestBase.java 2728 2015-05-14 20:00:03Z CHTAKOS $
 */
abstract public class CommonUmlTestBase {

	/** Empty in-memory configuration. */
	protected Config emptyCfg;

	/** Empty model (with its default null-packages and null-classes). */
	protected UmlModel emptyModel;

	@Before
	public void setUp() throws Exception {
		emptyModel = createEmptyModel();
		emptyCfg = emptyModel.getCfg();
	}

	@After
	public void tearDown() {
		emptyModel.clear();
	}

	/**
	 * Returns an empty model with empty configuration. Useful for tests where you want to check
	 * that the elements belong to the same model, so you can use {@link #emptyModel} as usual, and
	 * create another empty model with this method.
	 * 
	 * @throws ApplicationException
	 */
	protected final UmlModel createEmptyModel() throws ApplicationException {
		Properties props = new Properties();
		// props.put(Config.KEY_MODEL_FILENAME, "inexisting.eap");
		// props.put(Config.KEY_VALIDATION_SCOPE, "WG14");
		return new UmlModel(new Config(props, null));
	}

	/**
	 * Creates an empty model configured with a non-CIM (i.e., non-default) model package with given
	 * name.
	 * 
	 * @throws ApplicationException
	 */
	protected final static UmlModel createModelWithNonCimNature(String nonCimModelName)
			throws ApplicationException {
		Properties props = new Properties();
		props.put(Config.KEY_MODEL_NATURE_IEC61850, nonCimModelName);
		props.put(Config.KEY_VALIDATION_IEC61850_PACKAGE72_TOP, "IEC61850_7_2");
		return new UmlModel(new Config(props, null));
	}

	/**
	 * Creates a model configured with a non-CIM (i.e., non-default) model package containing a
	 * single package, and returns that one package. This is utility method.
	 * 
	 * @throws ApplicationException
	 */
	protected final static UmlPackage createNonCimPackage(String nonCimPckName)
			throws ApplicationException {
		UmlModel model = createModelWithNonCimNature(nonCimPckName);
		UmlPackage mp = UmlPackage.basic(model, nonCimPckName);
		assertSame(Nature.IEC61850, mp.getNature());
		return mp;
	}

	static void createPresenceConditionsEnumType(UmlPackage mp, String... literals) {
		UmlClass pcEnum = UmlClass.basic(mp, "PresenceConditions", "enumeration", "cond");
		assertTrue("is presence condition class", pcEnum.isConditionEnumeration());
		for (String lit : literals) {
			pcEnum.addAttribute(null, lit, "enum");
		}
	}

	/** For testing only: Adds association with default data. */
	static UmlAssociation addAssociation(UmlClass clazz1, UmlClass clazz2) {
		UmlAssociationEnd source = UmlAssociationEnd.basic(clazz1, "C1");
		UmlAssociationEnd target = UmlAssociationEnd.basic(clazz2, "C2");
		return clazz1.addAssociation(source, target, new UmlObjectData(""),
				UmlAssociation.Data.empty());
	}

	/** For testing only: Adds informative association with default data. */
	static UmlAssociation addInfAssociation(UmlClass clazz1, UmlClass clazz2) {
		UmlAssociationEnd source = UmlAssociationEnd.basic(clazz1, "C1");
		UmlAssociationEnd target = UmlAssociationEnd.basic(clazz2, "C2");
		return clazz1.addAssociation(source, target, new UmlObjectData(null, "", new UmlStereotype(
				UmlStereotype.INFORMATIVE)), UmlAssociation.Data.empty());
	}

	static UmlClass createClassWithAttributes(UmlPackage mp, String className, String... attr) {
		UmlClass c = UmlClass.basic(mp, className, "");
		for (String a : attr) {
			UmlClass type = UmlClass.basic(mp, String.format("%sType", a), "");
			c.addAttribute(type, a, "");
		}
		return c;
	}
}
